#ifndef curve25519_sandy2x_H
#define curve25519_sandy2x_H

#include "crypto_scalarmult_curve25519.h"

extern struct crypto_scalarmult_curve25519_implementation
    crypto_scalarmult_curve25519_sandy2x_implementation;

#endif
