#ifndef ladder_H
#define ladder_H

#ifdef __cplusplus
extern "C" {
#endif

#include "fe.h"
#include "ladder_namespace.h"

extern void ladder(fe *, const unsigned char *);

#ifdef __cplusplus
}
#endif

#endif /* ifndef ladder_H */

